#Copyright (c) 2015 3D Control Systems LTD

#3DPrinterOS client is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#3DPrinterOS client is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with 3DPrinterOS client.  If not, see <http://www.gnu.org/licenses/>.

# Author: Vladimir Avdeev <another.vic@yandex.ru>, Oleg Panasevych <panasevychol@gmail.com>

import os
import json
import threading
import collections

import paths

REMOTE_IP = ""
LOCAL_IP = "127.0.0.1"

error_report_lock = threading.Lock()

def get_settings():
    return Config.instance().settings

def get_profiles():
    return Config.instance().profiles

def get_app():
    return Config.instance().app

def get_error_reports():
    return Config.instance().error_reports

def create_error_report(code, message, usb_info=None, logger=None, is_blocking=None):
    error_report = {}
    error_report['code'] = code
    error_report['message'] = message
    error_report['usb_info'] = usb_info
    error_report['is_blocking'] = is_blocking
    with error_report_lock:
        if not error_report in Config.instance().error_reports:
            Config.instance().error_reports.append(error_report)
    if logger:
        message = "Code=%d [ %s ]" % (code, message)
        if is_blocking:
            logger.warning("! " + message)
        else:
            logger.info("* " + message)


class Singleton(object):
    lock = threading.Lock()
    _instance = None

    @classmethod
    def instance(cls):
        with cls.lock:
            if not cls._instance:
                print "Creating new instance of " + cls.__name__
                cls._instance = cls()
        return cls._instance

def update_dict(base_dict, new_dict):
    updated_dict = {}
    updated_dict.update(base_dict)
    for key, value in new_dict.items():
        value_to_update = updated_dict.get(key)
        if type(value_to_update) != type(value):
            updated_dict[key] = value
        elif isinstance(value_to_update, dict) and isinstance(value, dict):
            updated_dict[key] = update_dict(value_to_update, value)
    return updated_dict


class Config(Singleton):

    DEFAULT_SETTINGS_FILE = os.path.join(os.path.dirname(__file__), 'default_settings.json')
    USER_SETTINGS_FILE = os.path.join(paths.current_settings_folder(), 'user_settings.json')

    def __init__(self):
        self.init_settings()
        self.profiles = None
        self.app = None
        self.error_reports = collections.deque()

    def init_settings(self):
        initial_settings = self.load_settings(self.USER_SETTINGS_FILE, warnings = False)
        settings = self.load_settings(self.DEFAULT_SETTINGS_FILE)
        if initial_settings:
            settings = update_dict(initial_settings, settings)
        if settings != initial_settings:
            self.save_settings(settings, self.USER_SETTINGS_FILE)
        self.settings = settings

    def load_settings(self, settings_path, warnings=True):
        try:
            with open(settings_path) as settings_file:
                config = json.loads(settings_file.read())
        except Exception as e:
            if warnings:
                print "Error reading config from %s: %s" % (settings_path, str(e))
        else:
            return config

    def save_settings(self, settings, settings_path=None):
        if not settings_path:
            settings_path=self.USER_SETTINGS_FILE
        try:
            with open(settings_path, 'w') as settings_file:
                json_config = json.dumps(settings, sort_keys = True, indent = 4, separators = (',', ': '))
                settings_file.write(json_config)
        except Exception as e:
            print "Error writing config to %s: %s" % (settings_path, str(e))
        else:
            self.settings = settings

    def restore_default_settings(self):
        if os.path.exists(self.USER_SETTINGS_FILE):
            try:
                os.remove(self.USER_SETTINGS_FILE)
            except Exception as e:
                return str(e)
        self.settings = self.load_settings(self.DEFAULT_SETTINGS_FILE)

    def set_profiles(self, profiles):
        self.profiles = profiles

    def set_app_pointer(self, app):
        self.app = app

