import time
import json
import socket
import logging
import threading

import config
import usb_detect

LISTEN_TIMEOUT = 3

class Listening(threading.Thread):

    def __init__(self, port):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.data = None
        self.addr = None
        self.socket = self.create_socket(port)
        threading.Thread.__init__(self)

    def create_socket(self, port):
        self.logger.debug("Creating listen socket for port %d" % port)
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.bind(('', port))
            sock.settimeout(LISTEN_TIMEOUT)
        except socket.error as e:
            self.logger.debug("Socket init error. Port:" + str(port), exc_info=True)
        except ValueError as e:
            self.logger.debug("Not valid port number:" + str(port), exc_info=True)
        else:
            self.logger.debug("...done")
            return sock

    def run(self):
        try:
            self.data, self.addr = self.socket.recvfrom(1024)
        except socket.error:
            pass

class NetPrinterScan:

    RETRIES = 1
    RETRY_TIMEOUT = 0.1

    def __init__(self, profile):
        self.logger = logging.getLogger('app.' + __name__)
        self.profile = profile
        network_profile = profile['network_detect']
        self.listen_port = network_profile.get('listen_port', None)
        self.broadcast_port = network_profile.get('broadcast_port', None)
        self.target_port = network_profile.get('target_port', None)
        message = network_profile.get('message', None)
        if network_profile.get('json_message', None):
            message = json.dumps(message)
        self.message = message
        self.already_detected = []
        self.discovered_printers = []
        if self.broadcast_port and self.listen_port and self.target_port:
            self.listen = Listening(self.listen_port)
            self.listen.start()
            self.broadcast()
        else:
            error = "Error in config file. Section network_detect of profile: " + str(self.profile)
            self.logger.critical(error)

    def create_broadcasting_socket(self):
        self.logger.debug("Creating broadcast socket on port %d" % self.broadcast_port)
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            #bc_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind(('', self.broadcast_port))
        except socket.error as e:
            self.logger.debug("Socket init error. Port: %d" % self.broadcast_port, exc_info=True)
        except ValueError as e:
            self.logger.debug("Not valid port number: %d" % self.broadcast_port, exc_info=True)
        else:
            return sock

    def broadcast(self):
        self.logger.debug("Sending broadcast to port %d" % self.target_port)
        bc_socket = self.create_broadcasting_socket()
        if bc_socket:
            attempt = 1
            while attempt < self.RETRIES + 1:
                try:
                    bc_socket.sendto(str(self.message), ('255.255.255.255', self.target_port))
                except socket.error as e:
                    self.logger.debug("Error sending broadcast", exc_info=True)
                    #self.logger.debug("Timeout on port:" + str(self.listen_port))
                attempt += 1
                time.sleep(self.RETRY_TIMEOUT)
                self.listen.join()
                if self.listen.addr:
                    self.process_response(self.listen.data, self.listen.addr)
                    break
            self.logger.debug("Done listening to port " + str(self.listen_port))
            bc_socket.close()

    def process_response(self, response, addr):
        if addr not in self.already_detected:
            self.already_detected.append(addr)
            printer = {'IP': addr[0], 'port': addr[1]}
            if self.profile['network_detect']['json_response']:
                try:
                    response = json.loads(response)
                except:
                    self.logger.debug('Response from printer should be valid json. Its malformed or not json')
                else:
                    ip_field = self.profile['network_detect']['IP_field']
                    iserial_field = self.profile['network_detect']['SNR_field']
                    vid_field = self.profile['network_detect']['VID_field']
                    pid_field = self.profile['network_detect']['PID_field']
                    printer['IP'] = response[ip_field]
                    if addr[0] != printer['IP']:
                        self.logger.warning("Detected printer IP didn't match with IP field in response")
                    printer['SNR'] = response[iserial_field]
                    printer['VID'] = usb_detect.USBDetector.format_vid_or_pid(response[vid_field])
                    printer['PID'] = usb_detect.USBDetector.format_vid_or_pid(response[pid_field])
            self.discovered_printers.append(printer)

class NetworkDetector:

    def __init__(self):
        self.profiles = config.get_profiles()

    def get_printers(self):
        logger = logging.getLogger(self.__class__.__name__)
        logger.info('Scanning for network printers...')
        printers = []
        for profile in self.profiles:
            if 'network_detect' in profile:
                scanner = NetPrinterScan(profile)
                printers.extend(scanner.discovered_printers)
        logger.info('Discovered printers:\n' + str(printers))
        return printers

if __name__ == '__main__':
    class FakeApp:
        pass
    fa = FakeApp()
    fa.stop_flag = False
    import user_login
    import log
    log.create_logger("", None)
    user_login = user_login.UserLogin(fa)
    user_login.wait()
    if user_login.user_token and hasattr(user_login, "profiles"):
        config.Config.instance().set_profiles(user_login.profiles)
    detector = NetworkDetector()
    printers = detector.get_printers()
    print "Detected network printers: "
    print json.dumps(printers)


